@@ -1,16 +1,17 @@
-import io
 import os
 import pathlib
+import shutil
 import sys
 import tarfile
 import tempfile
 import timeit
 from datetime import datetime
 from logging import Logger
-from typing import Dict, Optional, List
+from typing import Dict, Optional, List, Set, Tuple
 
-from docker.models.containers import Container
+import docker
 from docker.errors import NotFound
+from docker.models.containers import Container
 
 from cincan.command_log import FileLog, read_with_hash
 from cincan.file_tool import FileMatcher
@@ -36,34 +37,53 @@ class TarTool:
         if not self.explicit_file and not upload_files:
             return  # nothing to upload
 
-        if self.explicit_file:
+        if self.explicit_file == '-':
+            # tar from stdin
+            tar_file = tempfile.TemporaryFile()
+            try:
+                shutil.copyfileobj(sys.stdin.buffer, tar_file)
+                tar_file.seek(0)
+                self.__list_members(tar_file, in_files)
+                tar_file.seek(0)
+                self.__put_archive(tar_file)
+            finally:
+                tar_file.close()  # should be deleted by close
+        elif self.explicit_file:
             # tar file provided, use it as-it-is
-            if self.explicit_file == '-':
-                tar_content = sys.stdin.buffer.read()
-            else:
-                explicit_file = pathlib.Path(self.explicit_file)
-                with explicit_file.open("rb") as f:
-                    tar_content = f.read()
-
-            with tarfile.open(fileobj=io.BytesIO(tar_content), mode="r") as tar:
-                for m in tar.getmembers():
-                    self.logger.debug(f"checking in file {m.name}")
-                    self.upload_stats[m.name] = [m.size, m.mtime]  # [size, mtime]
-                    if not m.isfile():
-                        continue
-                    m_fileobj = tar.extractfile(m)
-                    m_file = pathlib.Path(m.name)
-                    m_md = read_with_hash(m_fileobj.read)
-                    in_files.append(
-                        FileLog(m_file.resolve(), m_md, datetime.fromtimestamp(m.mtime)))
+            explicit_path = pathlib.Path(self.explicit_file)
+            with explicit_path.open("rb") as tar_file:
+                self.__list_members(tar_file, in_files)
+            with explicit_path.open("rb") as tar_file:
+                self.__put_archive(tar_file)
         else:
-            tar_content = self.__create_tar(upload_files, in_files)
+            # collect a tar file, and upload it
+            tar_file = self.__create_tar(upload_files, in_files)
+            try:
+                self.__put_archive(tar_file)
+            finally:
+                tar_file.close()
+
+    def __put_archive(self, tar_content):
         put_arc_start = timeit.default_timer()
         self.container.put_archive(path=self.work_dir, data=tar_content)
         self.logger.debug("put_archive time %.4f s", timeit.default_timer() - put_arc_start)
 
-    def __create_tar(self, upload_files: Dict[pathlib.Path, str], in_files: List[FileLog]) -> bytes:
-        file_out = io.BytesIO()
+    def __list_members(self, tar_content, in_files: List[FileLog]):
+        with tarfile.open(fileobj=tar_content, mode="r") as tar:
+            for m in tar.getmembers():
+                self.logger.debug(f"checking in file {m.name}")
+                # file size, modification time, upload time
+                self.upload_stats[m.name] = [m.size, m.mtime, datetime.now().timestamp()]
+                if not m.isfile():
+                    continue
+                m_fileobj = tar.extractfile(m)
+                m_file = pathlib.Path(m.name)
+                m_md = read_with_hash(m_fileobj.read)
+                in_files.append(
+                    FileLog(m_file.resolve(), m_md, datetime.fromtimestamp(m.mtime)))
+
+    def __create_tar(self, upload_files: Dict[pathlib.Path, str], in_files: List[FileLog]):
+        file_out = tempfile.TemporaryFile()
         tar = tarfile.open(mode="w", fileobj=file_out)
 
         # need to have all directories explicitly, otherwise seen them to be created with root
@@ -87,7 +107,8 @@ class TarTool:
             else:
                 tar_file = tar.gettarinfo(host_file, arcname=arc_name)
                 tar_file.mode = 511  # 777 - allow all to access (uid may be different in container)
-                self.upload_stats[arc_name] = [tar_file.size, tar_file.mtime]  # [size, mtime]
+                # file size, modification time, upload time
+                self.upload_stats[arc_name] = [tar_file.size, tar_file.mtime, datetime.now().timestamp()]
                 if host_file.is_file():
                     # put file to tar
                     with host_file.open("rb") as f:
@@ -102,8 +123,8 @@ class TarTool:
                     tar.addfile(tar_file)
                 else:
                     raise Exception(f"Cannot upload file of unknown type {arc_name}")
-        tar.close()
-        return file_out.getvalue()
+        file_out.seek(0)
+        return file_out
 
     @classmethod
     def __new_directory(cls, name: str) -> tarfile.TarInfo:
@@ -114,9 +135,9 @@ class TarTool:
         p_file.mode = 511  # 777 - allow all to access (uid may be different in container)
         return p_file
 
-    def __read_single_file(self, filepath: pathlib.Path, skip_comment: bool = False) -> List[str]:
-        """Method for reading contents single file from the container.
-        Return list of strings, where one index represents single line of the file.
+    def __read_config_file(self, filepath: pathlib.Path, skip_comment: bool = False) -> List[str]:
+        """Read configuration file.
+        Returns a list of strings, where one index represents single line of the file.
         """
 
         file_lines = []
@@ -143,23 +164,77 @@ class TarTool:
             self.logger.debug(e)
         return file_lines
 
-    def download_files(self, output_filters: List[FileMatcher] = None, no_defaults: bool = False) -> List[FileLog]:
+    def download_files(self, filters: List[FileMatcher] = None, no_defaults: bool = False,
+                       file_paths: List[str] = None, implicit_output=True) -> List[FileLog]:
+        """Download modified files, filtered as required"""
+        # check all modified (includes the ones we uploaded)
+        candidates = sorted(
+            [d['Path'] for d in filter(lambda f: 'Path' in f, self.container.diff() or [])], reverse=True)
+        # note: candidates start with / as path container absolute
+        candidates = self.__filter_files(candidates, filters, no_defaults)
 
+        # write to a tar?
+        explicit_file = None
+        try:
+            if self.explicit_file == '-':
+                # write tar to stdout
+                explicit_file = tarfile.open(mode="w|", fileobj=sys.stdout.buffer)
+            elif self.explicit_file:
+                # write tar file
+                explicit_file = tarfile.open(self.explicit_file, "w")
+
+            files_to_do = set(candidates)
+            out_files = []
+
+            if implicit_output and self.work_dir != '/':
+                # container has non-root working directory, get it at once!
+                self.logger.debug("%d files to download, looking from %s...", len(files_to_do), self.work_dir)
+                log = self.__download_file_set(self.work_dir, files_to_do, write_to=explicit_file)
+                out_files.extend(log)
+
+            # explicit result directories
+            for fp in file_paths or []:
+                if not files_to_do:
+                    continue
+                fp_in_cont = (pathlib.Path(self.work_dir) / fp).as_posix()
+                self.logger.debug("%d files to download, looking from %s...", len(files_to_do), fp_in_cont)
+                log = self.__download_file_set(fp_in_cont, files_to_do, write_to=explicit_file)
+                out_files.extend(log)
+
+            if implicit_output and files_to_do:
+                # go for each missing file individually
+                self.logger.debug("%d files to download, fetching each individually", len(files_to_do))
+                files_to_load = sorted(files_to_do)
+                for f in files_to_load:
+                    # separately download and possibly copy remaining result files
+                    log = self.__download_file_set(f, files_to_do, write_to=explicit_file)
+                    out_files.extend(log)
+
+            if files_to_do:
+                self.logger.debug("%d files in diff not downloaded:", len(files_to_do))
+                for f in sorted(files_to_do):
+                    self.logger.debug("  %s", f)
+            return out_files
+        finally:
+            explicit_file and explicit_file.close()
+
+    def __filter_files(self, candidates: List[str], filters: List[FileMatcher] = None,
+                       no_defaults: bool = False) -> List[str]:
+        """Filter list of candidate files to download"""
         # Sort by excluding and including filters
         # Including filter has more power than excluding one!
-        output_filters_to_exclude = output_filters.copy() if output_filters else []
-        output_filters_to_include = [output_filters_to_exclude.pop(i) for i, f in enumerate(output_filters_to_exclude)
-                                     if f.include] if output_filters else []
+        output_filters_to_exclude = filters.copy() if filters else []
+        output_filters_to_include = [output_filters_to_exclude.pop(i) for i, f in
+                                     enumerate(output_filters_to_exclude)
+                                     if f.include] if filters else []
 
         # Check if container has .cincanignore file - these are not downloaded by default
         ignore_file = pathlib.Path(self.work_dir) / IGNORE_FILENAME
-        ignore_paths = self.__read_single_file(ignore_file, skip_comment=True)
+        ignore_paths = self.__read_config_file(ignore_file, skip_comment=True)
         # Ignore the ignorefile itself..
         ignore_paths.append(IGNORE_FILENAME)
-        # check all modified (includes the ones we uploaded)
-        candidates = sorted([d['Path'] for d in filter(lambda f: 'Path' in f, self.container.diff() or [])],
-                            reverse=True)
         # remove files which are paths to files
+        candidates = candidates.copy()
         skip_set = set()
         for i, c in enumerate(candidates):
             if c in skip_set:
@@ -189,16 +264,16 @@ class TarTool:
                 ignore_filters.append(FileMatcher(file, include=False))
 
         # If user has not defined output_filters, use .cincanignore from container if not set to be ignored
-        if not output_filters and ignore_paths and not no_defaults:
+        if not filters and ignore_paths and not no_defaults:
             self.logger.debug("No user provided output filters - using .cincanignore")
             for filth in ignore_filters or []:
                 candidates = filth.filter_download_files(candidates, self.work_dir)
 
-        elif output_filters and ignore_paths:
+        elif filters and ignore_paths:
 
             # Check if user has defined to not use container specific output filters
             if no_defaults:
-                for filth in output_filters or []:
+                for filth in filters or []:
                     candidates = filth.filter_download_files(candidates, self.work_dir)
             elif output_filters_to_include:
                 # If we have some including filters, only those are applied
@@ -220,49 +295,54 @@ class TarTool:
                     candidates = filth.filter_download_files(candidates, self.work_dir)
 
         else:
-            for filth in output_filters or []:
+            for filth in filters or []:
                 # remove non-matching files
                 candidates = filth.filter_download_files(candidates, self.work_dir)
+        return candidates
 
-        # write to a tar?
-        explicit_file = None
-        try:
-            if self.explicit_file == '-':
-                explicit_file = tarfile.open(mode="w|", fileobj=sys.stdout.buffer)
-            elif self.explicit_file:
-                explicit_file = tarfile.open(self.explicit_file, "w")
-            out_files = []
-            for f in candidates:
-                log = self.__download_file_maybe(f, write_to=explicit_file)
-                out_files.extend(log)
-            return out_files
-        finally:
-            explicit_file and explicit_file.close()
+    def __download_file_set(self, file_path: str, files: Set[str],
+                            write_to: Optional[tarfile.TarFile] = None) -> List[FileLog]:
+        """Download files by a path, copy matching files into host"""
+        base_path = pathlib.Path(file_path)
 
-    def __download_file_maybe(self, file_name: str, write_to: Optional[tarfile.TarFile] = None) -> List[FileLog]:
-        host_file = pathlib.Path(
-            (file_name[len(self.work_dir):] if file_name.startswith(self.work_dir) else file_name))
-
-        # fetch the file from container in its own tar ball
+        # fetch the path from container in its own tar ball
         get_arc_start = timeit.default_timer()
-        chunks, stat = self.container.get_archive(file_name)
-        file_modified = self.__check_for_download(host_file, stat)
-        if not file_modified:
-            return []  # do not attempt download
+        try:
+            chunks, _ = self.container.get_archive(file_path)
+        except docker.errors.NotFound:
+            self.logger.debug("Not found in container: %s", file_path)
+            return []
 
         # read the tarball into temp file
         tmp_tar = tempfile.TemporaryFile()
         for c in chunks:
-            # self.logger.debug(f"chunk of {len(c)} bytes")
             tmp_tar.write(c)
-        self.logger.debug("get_archive time %.4f s", timeit.default_timer() - get_arc_start)
+        self.logger.debug("get_archive %s time %.4f s", file_path, timeit.default_timer() - get_arc_start)
 
         tmp_tar.seek(0)
         down_tar = tarfile.open(fileobj=tmp_tar, mode="r|")
         out_files = []
         for tar_file in down_tar:
-            # Note, we trust all intermediate directories to be provided in the tar files
-            local_file = host_file.parent / tar_file.name
+            file_in_cont = (base_path.parent or base_path) / tar_file.name
+            cont_full_name = file_in_cont.as_posix()
+            if cont_full_name not in files:
+                continue  # not interested in this
+            files.remove(cont_full_name)
+
+            if cont_full_name.startswith(self.work_dir):
+                file_in_host = pathlib.Path(cont_full_name[len(self.work_dir):])
+            elif cont_full_name.startswith('/'):
+                file_in_host = pathlib.Path(cont_full_name[1:])
+            else:
+                self.logger.warning("skipping file %s", cont_full_name)
+                continue
+            if file_in_host.is_absolute():
+                self.logger.warning("skipping suspicious file %s", cont_full_name)
+                continue
+            modified, unmodified = self.__check_if_modified(file_in_host, tar_file)
+            if unmodified:
+                continue  # no need to download
+
             md = ''
             timestamp = datetime.now()
             if write_to:
@@ -271,90 +351,91 @@ class TarTool:
                     tf_data = down_tar.extractfile(tar_file)
                     md = read_with_hash(tf_data.read, temp_file.write)
 
-                    write_tf = tarfile.TarInfo(local_file.as_posix())
+                    write_tf = tarfile.TarInfo(file_in_host.as_posix())
                     write_tf.mtime = tar_file.mtime
                     write_tf.mode = tar_file.mode
                     write_tf.size = temp_file.tell()
                     temp_file.seek(0)
                     write_to.addfile(write_tf, fileobj=temp_file)
-            elif local_file == host_file and tar_file.isfile():
-                # this is the file we were looking for
-                if not host_file.exists():
+            elif tar_file.isfile():
+                # this is a file we were looking for
+                if not file_in_host.exists():
                     # no local file or explicit output asked, this is too easy
-                    self.logger.info(f"=> {host_file.as_posix()}")
+                    self.logger.info(f"=> {file_in_host.as_posix()}")
                     tf_data = down_tar.extractfile(tar_file)
-                    if host_file.parent:
-                        host_file.parent.mkdir(parents=True, exist_ok=True)
-                    with host_file.open("wb") as f:
+                    if file_in_host.parent:
+                        file_in_host.parent.mkdir(parents=True, exist_ok=True)
+                    with file_in_host.open("wb") as f:
                         md = read_with_hash(tf_data.read, f.write)
                 else:
                     # compare by hash, if should override local file
                     tf_data = down_tar.extractfile(tar_file)
-                    temp_file = pathlib.Path(host_file.as_posix() + '_TEMP')
+                    temp_file = pathlib.Path(file_in_host.as_posix() + '_TEMP')
                     self.logger.debug(f"creating temp file {temp_file.as_posix()}")
 
                     # calculate hash for file from tar, copy it to temp file
-                    if host_file.parent:
-                        host_file.parent.mkdir(parents=True, exist_ok=True)
+                    if file_in_host.parent:
+                        file_in_host.parent.mkdir(parents=True, exist_ok=True)
                     with temp_file.open("wb") as f:
                         md = read_with_hash(tf_data.read, f.write)
 
-                    # calculate hash for existing file
-                    with host_file.open("rb") as f:
-                        host_digest = read_with_hash(f.read)
-
-                    self.logger.info(f"=> {host_file.as_posix()}")
-                    if md == host_digest:
-                        self.logger.debug(f"identical file {host_file.as_posix()} digest {md}, no action")
-                        # The file is identical as uploaded, but timestamps tell different story.
-                        # Assuming it was created identical, adding the log entry
-                        temp_file.unlink()
+                    self.logger.info(f"=> {file_in_host.as_posix()}")
+                    if modified:
+                        # modified by timestamp or size, overwrite
+                        file_in_host.unlink()
+                        temp_file.rename(file_in_host)
                     else:
-                        self.logger.debug(
-                            f"file {host_file.as_posix()} digest in container {md}, in host {host_digest}")
-                        host_file.unlink()
-                        temp_file.rename(host_file)
+                        # not sure if modified, calculate hash for existing file
+                        with file_in_host.open("rb") as f:
+                            host_digest = read_with_hash(f.read)
+
+                        if md == host_digest:
+                            self.logger.debug(f"identical file {file_in_host.as_posix()} digest {md}, no action")
+                            # The file is identical as uploaded
+                            # Assuming it was created identical, adding the log entry
+                            temp_file.unlink()
+                        else:
+                            # digest changed, modified despite identical size and timestamps
+                            self.logger.debug(
+                                f"file {file_in_host.as_posix()} digest in container {md}, in host {host_digest}")
+                            file_in_host.unlink()
+                            temp_file.rename(file_in_host)
             else:
-                if not host_file.exists():
+                if not file_in_host.exists():
                     # must be a directory we need
-                    self.logger.info(f"=> {local_file.as_posix()}/")
-                    local_file.mkdir(parents=True, exist_ok=True)
-                    timestamp = datetime.fromtimestamp(local_file.stat().st_mtime)
+                    self.logger.info(f"=> {file_in_host.as_posix()}/")
+                    file_in_host.mkdir(parents=True, exist_ok=True)
+                    timestamp = datetime.fromtimestamp(file_in_host.stat().st_mtime)
                 else:
                     pass  # no action required
-            out_files.append(FileLog(host_file.resolve(), md, timestamp))
+            out_files.append(FileLog(file_in_host.resolve(), md, timestamp))
         tmp_tar.close()
         return out_files
 
-    def __check_for_download(self, host_file: pathlib.Path, stat: Dict) -> bool:
-        """Check if file should be downloaded from the container"""
+    def __check_if_modified(self, host_file: pathlib.Path, file_info: tarfile.TarInfo) -> Tuple[bool, bool]:
+        """Check if file has been modified, or not modified in the container"""
         up_stat = self.upload_stats.get(host_file.as_posix())
+        if up_stat is None:
+            return True, False  # not uploaded, must be downloaded
+
+        # original size, upload time, original creation time
+        orig_size, orig_time, up_time = up_stat
+
         # NOTE: for directories (?) size from container is 4096 -> mismatch for directories!!!
-        if up_stat and 'size' in stat:
-            down_size = int(stat['size'])
-            up_size = up_stat[0]
-            up_down_size_mismatch = up_size != down_size
-            if up_down_size_mismatch:
-                self.logger.debug(f"size {host_file.as_posix()} change {up_size} -> {down_size}")
-                return True
+        down_size = file_info.size
+        if orig_size != down_size:
+            self.logger.debug(f"size {host_file.as_posix()} change {orig_size} -> {down_size}")
+            return True, False
 
-        if up_stat and 'mtime' in stat:
-            # MacOS needs some timezone specifications
-            if sys.platform == "darwin":
-                up_time = datetime.utcfromtimestamp(int(up_stat[1]))
-            else:
-                up_time = datetime.fromtimestamp(int(up_stat[1]))
-            down_time_s = stat['mtime']  # seconds + timezone
-            up_time_s = up_time.strftime(self.time_format_seconds)  # seconds
-            time_now_s = datetime.now().strftime(self.time_format_seconds)
-            if down_time_s.startswith(up_time_s):
-                # looks like timestamp not updated, but down_time is seconds and up_time has more precision
-                if up_time_s == time_now_s:
-                    self.logger.debug(f"timestamps {host_file.as_posix()} now {down_time_s}, may or may not be updated")
-                    return True  # edited, but actually we do not know
-                self.logger.debug(f"timestamp {host_file.as_posix()} not updated {down_time_s}")
-                return False  # not edited, we are sure
-            self.logger.debug(f"timestamp {host_file.as_posix()} updated {up_time_s} -> {down_time_s}")
-            return True  # edited, we are sure
-
-        return True  # tell edited, but we have no idea
+        # remove fractions for comparison
+        down_time_s = int(file_info.mtime)
+        orig_time_s = int(orig_time)
+        up_time_s = int(up_time)
+        if orig_time_s != down_time_s:
+            self.logger.debug(f"{host_file.as_posix()} mtime updated at container")
+            return True, False
+        if orig_time_s != up_time_s:
+            self.logger.debug(f"{host_file.as_posix()} old and not modified")
+            return False, True
+        self.logger.debug(f"{host_file.as_posix()} timestamp nor size tells if modified")
+        return False, False
